#!/usr/bin/perl
#clustermerger5.0.plx
use warnings;
use strict;
use List::UtilsBy qw(extract_by);

#program is designed to take summarized CSV files of UCLUST or REMARK results and merge into a single CSV
#version 3.0 replaces the cluster writing functions with a simpler List:Utils module function to automatically splice out array elements matching a Grep function
#PROGRAM HISTORY:
	#VERSION1.0
		#version1.0 worked poorly, and had a lot of weird code. Version2.0 polishes this up into a clean program.
		#version1.0 also had the program taxing the machine hard by having to search through each .csv many times. Made for a lot of opening and closing, and overall poor performance. 
			#version2.0 accomplishes initial finding by grepping early and once the master .csv has been compiled, it will take very little time to go through and merge. 
			
		#version1.0 worked by creating a master .CSV and working through .csv greps at the same time
			#worked TERRIBLE
		#version2.0 compiles master .csv first, then greps for each header
			#initially more redundant because the lines aren't merged on the go, but should be easier to implement and run smoother. 
		#version1.0 also would search through every single cluster file created. SO BAD!!!!
			#version2.0 ignores this by referencing simply the master CSV file, rather than having to open up each of the cluster files. 
	#VERSION2.0
		#version2.0 replaced the multiple .csv's with a single MASTER csv file, then wrote out the contents to separate cluster files that could be edited. 
		#This version suffered significantly from read/write time, and additionally had to rewrite the MASTER file periodically to update which clusters had been merged
		#This version is somewhat less CPU intensive, and has the same memory requirements, but will get incrementally more RAM efficient as time passes due to condensation of the MASTER array.
		#Additionally, this version had an extra couple hundred lines to do the same thing less efficiently. 
			#List:Utils also allows for grepping and splicing at the same time, which halves the number of checks AND eliminates the redundancy of having to reference and rewrite the MASTER csv file.
	#VERSION3.0
		#version 3.0 replaced index files with simpler array, and implemented LIST:UTILS extract_by function. 
		#still had runtime issues during Phase 1; pretty slow because it has to go through each file for each header, and the read speed is lacking yet I can't keep everything in memory. 
	#VERSION4.0
		#version4.0 replaces the cumbersome read through each file for each header function with a simpler master array that uses the extract_by function to match any of the headers read from an incoming file
		#reduces the number of checks by a potentially very large amount due to condensing as it goes, and because it's checking many headers simultaneously.
			#unfortunately, this means it has to check the ENTIRE array (which may be quite massive), but it has the benefit of doing so many fewer times. 
		#Due to the simultaneous reading and merging, it removes Phase 2 and 3. 
		#because it uses headers grabbed from each query file and not from a reference, Version 4 also removes the need for a header reference file. 
	#VERSION5.0
		#Version5.0 attempts to address the runtime issues with Version 4 by eliminating the multiple input csvs
		#the independent cluster CSV's must be pre-processed by removing all duplicates and leaving ONLY unique clusters
		#functionality of version 5.0 will be similar to version 3 (after master fasta generation), but use extract_by function of version 4.
			#will reinstate the need for a header reference file
			#program will run through, header by header, and condense array elements as it goes. 
			#runtime should improve as it works through the file. 
		#current version requires "remove duplicates" processing in Excel. 
		
	#for future versions, during file reformatting just keep track of an index value unique to each line
		#Replace species and compartment with single letter ID, concatenate with digits
		#as files reformat, use UNIQUE function to write OUT unique values to the write file; ignore writing duplicates. 
	
	
print "\n";
print "\t\t\tRunning perl version $^V\n" ;
print "\t\t\tProgram Developed by Ryan Christian\n";
print "\t\t\tVersion 1.0 developed 9-2-2016\n";
print "\t\t\tOriginal coding: 9-2-2017\n";
print "\t\t\tVersion 2.0 developed 2-21-2017\n";
print "\t\t\tVersion 3.0 developed 3-9-2017\n";
print "\t\t\tVersion 4.0 developed 3-10-2017\n";
print "\t\t\tVersion 5.0 developed 3-11-2017\n";
print "\t\t\tLast Modified: 3-11-2017\n";
print "\n\n";
		
print "Clustermerger5.0.plx is a tool designed to merge results of multiple UCLUST or ReMark outputs (summarized in .csv format)\n";
print "\tNOTE: this program requires a csv file of UNIQUE clusters only, and a reference file with constituent fasta headers\n";
print "For other information, please see program notes\n";		
#=======================================================================================================================================================================
#													PHASE 0: Setup
#=======================================================================================================================================================================
		
		
print "What species are in this analysis?\n";
print "Please input each 5-letter species abbreviation separated by spaces\n";
my $speciesinput = <STDIN>;
chomp $speciesinput;
my @species = split/ /, $speciesinput;
		
print "Please input CSV file containing ONLY unique clusters: \n";
my $input = <STDIN>;
chomp $input;
$input =~ s/"//g;

my $ctporprot;
if ($input =~ m/PROT/gi) {
	$ctporprot = "PROT";
	}	else {
		$ctporprot = "CTP";
}

print "Please specify writing directory: \n";
my $directory = <STDIN>;
chomp $directory;
$directory =~ s/"//g;
chdir $directory;

print "Please indicate header reference file: \n";
print "\tNote that file should ONLY contain headers, separated by newlines\n";
my $headerref = <STDIN>;
chomp $headerref;
$headerref =~ s/"//g;

print "Please indicate name for output file: \n";
my $master = <STDIN>;
chomp $master;
$master =~ s/"//g;
$master =~ s/.csv//g;
	#to avoid redundancy of ".csv.csv.csv" files, removes any instances of .csv
my $mastercsv = $master;
$mastercsv = "$mastercsv.csv";


my $rundeets = "$master-RunDetails.txt";
open (RUNDEETS, ">", $rundeets) or die "Couldn't open $rundeets for writing: $!";
print "Reading from cluster file $input\n";
print RUNDEETS "Analyzing species: ";

local $, = ",";
print RUNDEETS @species;
print "Analyzing species: ";
print @species;
print RUNDEETS "\n";

open (UNIQUECSV, "<", $input) or die "Couldn't open csv $input for reading: $!\n";
my @clusters = <UNIQUECSV>;
close UNIQUECSV;
chomp @clusters;
print "Finished reading clusters at " . localtime . "\n";


#=======================================================================================================================================================================
#													PHASE 1: Cluster Merging
#=======================================================================================================================================================================



print "Starting Merging at time " . localtime . "\n";
open (MASTEROUT, ">", $mastercsv) or die "couldn't open file $mastercsv for writing: $!\n";
open (HEADERS, "<", $headerref) or die "couldn't open header reference file $headerref for readng: $!\n";

my $singles = "$master-singletons.txt";
open (SINGLES, ">", $singles) or die "couldn't open singleton OUT file $singles for writing: $!\n";


my $currentspecies = '';
my @singletonclusters;
while (<HEADERS>) {
	if ($_ =~ m/^>/) {
		chomp $_;
		my $matchname = $_;
		(my $matchspecies) = $matchname =~ m/(\w{5})/;

		if ($currentspecies ne $matchspecies) {
			print "Finished merging headers from $currentspecies at " . localtime . "\n";
			print RUNDEETS "Finished merging headers from $currentspecies at " . localtime . "\n";
			print "Beginning merging headers from $matchspecies \n";
			print RUNDEETS "Beginning merging headers from $matchspecies \n";
			$currentspecies = $matchspecies;
		}
		
		my @matches = extract_by {m/($matchname)(,|$)/} @clusters;
			#NOTE FOR FUTURE:
			#extract_by DOES NOT ACCEPT AN ITERATOR VARIABLE!!!!
				#MUST assign another variable the value of the iterator and THEN use it in extract_by.... grrrrr
		my $matchnum = scalar @matches;
		
		print RUNDEETS "Merged $matchnum clusters for header $_\n";
		
		my @localheaders;
		foreach my $matchcluster (@matches) {
			my @splits = split/,/,$matchcluster;
			my @headers = grep/>\w{9}\d*/, @splits;
			chomp @headers;
			push @localheaders, @headers;
		}
		@localheaders = UNIQUE(@localheaders);
		my $headernum = scalar @localheaders;
		
		
		if ($headernum == 1) {
			local $, = ",";
			print SINGLES @localheaders;
			print SINGLES "\n";
			push @singletonclusters, @localheaders;
				#note: for huge numbers of singletons, this will likely drastically increase runtime. 
		} else {
			sort @localheaders;
				#condense all matched clusters to single group of headers
			local $, = ",";
			my $joined = join(',', @localheaders);
			push @clusters, $joined;
		}
	} else {
		next;
	} 
}
local $, = "\n";
print MASTEROUT @singletonclusters;
print MASTEROUT @clusters;

print "Finished merging clusters at ". localtime . "\n";
print RUNDEETS "Finished merging clusters at ". localtime . "\n";
close MASTEROUT;

#=======================================================================================================================================================================
#													PHASE 2: Stats and Write Merged File
#=======================================================================================================================================================================
#Beginning uclustcondenser part of the program
	#condenses down from the 1000's of clusters into a single .csv with sequence stats
	
my @tableheader = ("Cluster Name", "Cluster Population");
foreach my $myspecie (@species) {
	chomp $myspecie;
	push @tableheader, $myspecie;
}

for (my $i = 1; $i <= 500; $i++) {
push @tableheader, "Seq$i";
}
my $header = join(',', @tableheader);
@tableheader = '';
chomp $header;

open (MASTER, "<", $mastercsv) or die "couldn't open master CSV $mastercsv for reading: $!\n";

my $masterstats = "$master-STATS.csv";
open (MASTERSTATS, ">", $masterstats) or die "couldn't open stats CSV $masterstats for writing: $!\n";


print MASTERSTATS $header;
print MASTERSTATS "\n";

while (<MASTER>) {
	chomp $_;
	my @lineheaders = split/,/, $_;
	my @statistics = STATS (\@lineheaders, \@species);
	local $, = ",";
	print MASTERSTATS @statistics;
	print MASTERSTATS ",";
	print MASTERSTATS @lineheaders;
	print MASTERSTATS "\n";
}	

close MASTER;
close MASTERSTATS;
print "Finished writing clusters at ". localtime . "\n";
#=======================================================================================================================================================================
#														Subroutines
#=======================================================================================================================================================================






#SUBS!!!
sub UNIQUE {
	my %seen;
	#print "my initial array is " . scalar @_ . "elements long\n\n";
    grep !$seen{$_}++, @_;
	my @unique = keys (%seen);
	#print "my final array is " . scalar @unique . "elements long\n\n";
	my $scalar = scalar @_;
	return @unique;
		
#	print @unique;
#	print "\n\n";
		#should concatenate @headermatch and @localheaders and read @headermatch first, then pick through @localheaders for unique sequences.
}
		
		
sub STATS {
	#sub counts the number of headers, and counts the number of matches for each species. 
	my ($headers_ref, $species_ref) = @_;
	my @headers = @{ $headers_ref };
	my @species = @{ $species_ref };
		#dereference arrays
	
	#change local delimiter to a comma (CSV format)
	my @numbers;
	push @numbers, (scalar @headers);

	
	#does a grep for each species' header ID, and pushes the counted number to "@numbers"
	foreach my $specie (@species) {
		my @grep = grep (m/>$specie/g, @headers);
		
		#tests for array elements in @grep and returns number of matches OR "0" if undefined
		if (@grep) {
			push @numbers, (scalar @grep);
		} else {
			push @numbers, "0";
		}
}		
return @numbers;
}